@@ -8,8 +8,6 @@ use pyo3::types::{PyList, PyTuple};
 use serde::ser::SerializeSeq;
 use serde::{Serialize, Serializer};
 
-use crate::lookup_key::{LookupPath, PathItem};
-
 /// Used to store individual items of the error location, e.g. a string for key/field names
 /// or a number for array indices.
 #[derive(Clone, Eq, PartialEq, IntoPyObjectRef)]
@@ -71,20 +69,6 @@ impl From<usize> for LocItem {
     }
 }
 
-/// eventually it might be good to combine PathItem and LocItem
-impl From<PathItem> for LocItem {
-    fn from(path_item: PathItem) -> Self {
-        match path_item {
-            PathItem::S(s, _) => s.into(),
-            PathItem::Pos(val) => val.into(),
-            PathItem::Neg(val) => {
-                let neg_value = -(val as i64);
-                neg_value.into()
-            }
-        }
-    }
-}
-
 impl Serialize for LocItem {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
@@ -137,17 +121,6 @@ impl<'py> IntoPyObject<'py> for &'_ Location {
     }
 }
 
-impl From<&LookupPath> for Location {
-    fn from(lookup_path: &LookupPath) -> Self {
-        let v = lookup_path
-            .iter()
-            .rev()
-            .map(|path_item| path_item.clone().into())
-            .collect();
-        Self::List(v)
-    }
-}
-
 impl fmt::Display for Location {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match self {
